"""
Copyright (c) 2017, Jairus Martin.

Distributed under the terms of the MIT License.

The full license is in the file LICENSE.text, distributed with this software.

Created on Apr 16, 2017

@author: jrm
"""
from .html import *
from .md import Markdown
from .ipynb import Notebook
from .code import Code
from .raw import Raw



"""
Copyright (c) 2017, Jairus Martin.

Distributed under the terms of the MIT License.

The full license is in the file LICENSE.text, distributed with this software.

Created on Apr 17, 2017

@author: jrm
"""
from functools import lru_cache
from enaml.application import Application, ProxyResolver
from web.impl import lxml_components


class WebApplication(Application):
    """Base enaml web application that uses the widgets defined in
    `web.components.html`

    """

    def _default_resolver(self):
        return ProxyResolver(factories=lxml_components.FACTORIES)

    @lru_cache(1024)
    def resolve_proxy_class(self, declaration_class):
        """Cache the resolved value"""
        return super().resolve_proxy_class(declaration_class)



"""
Copyright (c) 2017-2019, Jairus Martin.

Distributed under the terms of the MIT License.

The full license is in the file LICENSE.text, distributed with this software.

Created on Aug 2, 2017

@author: jrm
"""
from atom.api import Instance
from pygments import lexers, highlight
from pygments.lexer import Lexer
from pygments.formatters import HtmlFormatter
from web.components.code import ProxyCode
from .lxml_raw import RawComponent, SourceType


class CodeComponent(RawComponent, ProxyCode):
    #: Lexer used
    lexer = Instance(Lexer)

    #: HTML Formatter
    formatter = Instance(HtmlFormatter)

    def _default_formatter(self):
        return HtmlFormatter(style=self.declaration.highlight_style)

    def _default_lexer(self):
        d = self.declaration
        if d.language:
            return lexers.find_lexer_class_by_name(d.language)()
        return lexers.guess_lexer(d.source)

    def set_source(self, source: SourceType):
        if isinstance(source, str):
            source = highlight(source, lexer=self.lexer, formatter=self.formatter)
        super().set_source(source)

    def set_language(self, language: str):
        d = self.declaration
        assert d is not None
        self.lexer = self._default_lexer()
        self.set_source(d.source)

    def set_highlight_style(self, style: str):
        d = self.declaration
        assert d is not None
        self.formatter = self._default_formatter()
        self.set_source(d.source)



"""
Copyright (c) 2017-2019, Jairus Martin.

Distributed under the terms of the MIT License.

The full license is in the file LICENSE.txt, distributed with this software.

Created on Apr 2, 2017

@author: jrm
"""
from __future__ import annotations

from typing import Any, Generator, Optional
from atom.api import (
    Atom,
    Event,
    Bool,
    Enum,
    ContainerList,
    Value,
    Int,
    Str,
    Dict,
    Instance,
    ForwardTyped,
    Typed,
    Coerced,
    observe,
    set_default,
)
from enaml.core.declarative import d_, Declarative
from enaml.widgets.toolkit_object import ToolkitObject, ProxyToolkitObject

ChangeDict = dict[str, Any]

try:
    from web.core.speedups import gen_id
except ImportError as e:

    def gen_id(tag, id=id, mod=divmod):
        """Generate a short id for the tag"""
        number = id(tag)
        output = ""
        alpha = "0123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"
        while number:
            number, digit = mod(number, 59)
            output += alpha[digit]
        return output


class ProxyTag(ProxyToolkitObject):
    #: Reference to the declaration
    declaration = ForwardTyped(lambda: Tag)

    #: A cached reference to the root element.
    #: WARNING: If the root is changed this becomes invalid
    root = ForwardTyped(lambda: ProxyTag)

    def xpath(self, query: str, **kwargs) -> Generator[ProxyToolkitObject, None, None]:
        """Perform an xpath lookup on the node"""
        raise NotImplementedError

    def render(self, method: str = "html", encoding: str = "unicode", **kwargs) -> str:
        """Render the node and all children"""
        raise NotImplementedError

    def set_attribute(self, name: str, value: Any):
        raise NotImplementedError


class Tag(ToolkitObject):
    #: Reference to the proxy object
    proxy = Typed(ProxyTag)

    #: Object ID
    id = d_(Str()).tag(attr=False)

    #: Tag name
    tag = d_(Str()).tag(attr=False)

    #: CSS classes
    cls = d_(Instance((list, tuple, str))).tag(attr=False)

    #: CSS styles
    style = d_(Instance((dict, str))).tag(attr=False)

    #: Node text
    text = d_(Str()).tag(attr=False)

    #: Node tail text
    tail = d_(Str()).tag(attr=False)

    #: Alt attribute
    alt = d_(Str()).tag(attr=False)

    #: Custom attributes not explicitly defined
    attrs = d_(Dict()).tag(attr=False)

    #: JS onclick definition
    onclick = d_(Str()).tag(attr=False)

    #: Used to tell js to send click events back to the server
    clickable = d_(Coerced(bool)).tag(attr=False)

    #: Used to tell js to send drag events back to the server and sets the
    #: draggable attribute. Must be used with ondragstart.
    draggable = d_(Coerced(bool)).tag(attr=False)

    #: JS ondragstart definition
    ondragstart = d_(Str()).tag(attr=False)

    #: JS ondragover definition
    ondragover = d_(Str()).tag(attr=False)

    #: JS ondragend definition
    ondragend = d_(Str()).tag(attr=False)

    #: JS ondragenter definition
    ondragenter = d_(Str()).tag(attr=False)

    #: JS ondragleave definition
    ondragleave = d_(Str()).tag(attr=False)

    #: JS ondrop definition
    ondrop = d_(Str()).tag(attr=False)

    #: Event triggered on click
    clicked = d_(Event())

    #: Event triggered when a drag starts
    dragstart = d_(Event())

    #: Event triggered when a drag ends
    dragend = d_(Event())

    #: Event triggered when a node is dragged over a valid drop target
    dragenter = d_(Event(ToolkitObject))

    #: Event triggered when a node leaves a valid drop target
    dragleave = d_(Event(ToolkitObject))

    #: Event triggered when a drop occurs
    dropped = d_(Event(ToolkitObject))

    def _default_id(self):
        return gen_id(self)

    @observe(
        "id",
        "tag",
        "cls",
        "style",
        "text",
        "tail",
        "alt",
        "attrs",
        "clickable",
        "draggable",
        "onclick",
        "ondragstart",
        "ondragover",
        "ondragend",
        "ondragenter",
        "ondragleave",
        "ondrop",
    )
    def _update_proxy(self, change: ChangeDict):
        """Update the proxy widget when the Widget data changes.

        This also notifies the root that the dom has been modified.

        """
        if self.proxy_is_active and change["type"] == "update":
            name = change["name"]
            value = change["value"]
            proxy = self.proxy
            assert proxy is not None
            handler = getattr(proxy, f"set_{name}", None)
            if handler is not None:
                handler(value)
            else:
                proxy.set_attribute(name, value)
            root = proxy.root
            if root is not None and root.rendered:
                self._notify_modified(
                    root.declaration,
                    {
                        "id": self.id,
                        "type": change["type"],
                        "name": name,
                        "value": value,
                    },
                )

    # =========================================================================
    # Object API
    # =========================================================================

    def child_added(self, child: Declarative):
        super().child_added(child)
        if self.proxy_is_active and isinstance(child, Tag):
            proxy = self.proxy
            assert proxy is not None
            root = proxy.root
            assert root is not None
            if root.rendered:
                self._notify_modified(
                    root.declaration,
                    {
                        "id": self.id,
                        "type": "added",
                        "name": "children",
                        "value": child.render(),
                        "before": self._next_child_id(child),
                    },
                )

    def child_moved(self, child: Declarative):
        super().child_moved(child)
        if self.proxy_is_active and isinstance(child, Tag):
            proxy = self.proxy
            assert proxy is not None
            root = proxy.root
            assert root is not None
            if root.rendered and proxy.child_moved(child.proxy):
                self._notify_modified(
                    root.declaration,
                    {
                        "id": self.id,
                        "type": "moved",
                        "name": "children",
                        "value": child.id,
                        "before": self._next_child_id(child),
                    },
                )

    def child_removed(self, child: Declarative):
        """Handles the child removed event.

        This will generate a modified event indicating which child was removed.

        """
        super().child_removed(child)
        if self.proxy_is_active and isinstance(child, Tag):
            proxy = self.proxy
            assert proxy is not None
            root = proxy.root
            assert root is not None
            if root.rendered:
                self._notify_modified(
                    root.declaration,
                    {
                        "id": self.id,
                        "type": "removed",
                        "name": "children",
                        "value": child.id,
                    },
                )

    def _notify_modified(self, root: Optional[Tag], change: ChangeDict):
        """Trigger a modified event on the root node. Subclasses may override
        this to update change parameters if needed.

        """
        if root is not None:
            root.modified(change)

    def _next_child_id(self, child: Tag) -> Optional[str]:
        """Find the id of the node after this child."""
        # Indicate where it was added
        children = self.children
        i = children.index(child) + 1
        for c in children[i:]:
            if isinstance(c, Tag):  # Ignore pattern nodes
                return c.id
        return None

    # =========================================================================
    # Tag API
    # =========================================================================
    def xpath(self, query: str, **kwargs) -> list[Tag]:
        """Find nodes matching the given xpath query

        Parameters
        ----------
        query: str
            The xpath query to run
        kwargs: dict
            Parameters to xpath.

        Returns
        -------
        nodes: List[Tag]
            List of tags matching the xpath query.

        """
        proxy = self.proxy
        assert proxy is not None
        return [n.declaration for n in proxy.xpath(query, **kwargs)]

    def prepare(self, **kwargs: dict[str, Any]):
        """Prepare this node for rendering.

        This sets any attributes given, initializes and actives the proxy
        as needed.

        """
        for k, v in kwargs.items():
            setattr(self, k, v)
        if not self.is_initialized:
            self.initialize()
        if not self.proxy_is_active:
            self.activate_proxy()

    def render(
        self, render_options: Optional[dict] = None, **kwargs: dict[str, Any]
    ) -> str:
        """Render this tag and all children to a string.

        Parameters
        -------
        render_options: dict
            Options to pass to render
        kwargs: dict
            Attributes to set on the view

        Returns
        -------
        html: str
            The rendered html content of the node.

        """
        self.prepare(**kwargs)
        proxy = self.proxy
        assert proxy is not None
        if render_options:
            return proxy.render(**render_options)
        return proxy.render()


class Html(Tag):
    __slots__ = "__weakref__"

    #: Set the tag name
    tag = set_default("html")

    #: Dom modified event. This will fire when any child node is updated, added
    #: or removed. Observe this event to handle updating websockets.
    modified = d_(Event(dict), writable=False).tag(attr=False)

    def _default_tag(self):
        return "html"


class Head(Tag):
    #: Set the tag name
    tag = set_default("head")


class Body(Tag):
    #: Set the tag name
    tag = set_default("body")


class Title(Tag):
    #: Set the tag name
    tag = set_default("title")


class P(Tag):
    #: Set the tag name
    tag = set_default("p")


class H1(Tag):
    #: Set the tag name
    tag = set_default("h1")


class H2(Tag):
    #: Set the tag name
    tag = set_default("h2")


class H3(Tag):
    #: Set the tag name
    tag = set_default("h3")


class H4(Tag):
    #: Set the tag name
    tag = set_default("h4")


class H5(Tag):
    #: Set the tag name
    tag = set_default("h5")


class H6(Tag):
    #: Set the tag name
    tag = set_default("h6")


class Hr(Tag):
    #: Set the tag name
    tag = set_default("hr")


class Br(Tag):
    #: Set the tag name
    tag = set_default("br")


class Pre(Tag):
    #: Set the tag name
    tag = set_default("pre")


class Kbd(Tag):
    #: Set the tag name
    tag = set_default("kbd")


class Samp(Tag):
    #: Set the tag name
    tag = set_default("samp")


class Var(Tag):
    #: Set the tag name
    tag = set_default("var")


class Div(Tag):
    #: Set the tag name
    tag = set_default("div")


class Span(Tag):
    #: Set the tag name
    tag = set_default("span")


class A(Tag):
    #: Set the tag name
    tag = set_default("a")

    #: Set the url
    href = d_(Str())

    #: Set the target options
    target = d_(Enum("", "_blank", "_self", "_parent", "_top", "framename"))

    @observe("href", "target")
    def _update_proxy(self, change: ChangeDict):
        super()._update_proxy(change)


class B(Tag):
    #: Set the tag name
    tag = set_default("b")


class I(Tag):
    #: Set the tag name
    tag = set_default("i")


class Strong(Tag):
    #: Set the tag name
    tag = set_default("strong")


class Em(Tag):
    #: Set the tag name
    tag = set_default("em")


class Mark(Tag):
    #: Set the tag name
    tag = set_default("mark")


class Small(Tag):
    #: Set the tag name
    tag = set_default("small")


class Del(Tag):
    #: Set the tag name
    tag = set_default("del")


class Ins(Tag):
    #: Set the tag name
    tag = set_default("ins")


class Sub(Tag):
    #: Set the tag name
    tag = set_default("sub")


class Sup(Tag):
    #: Set the tag name
    tag = set_default("sup")


class Q(Tag):
    #: Set the tag name
    tag = set_default("q")


class Blockquote(Tag):
    #: Set the tag name
    tag = set_default("blockquote")

    cite = d_(Str())

    @observe("cite")
    def _update_proxy(self, change: ChangeDict):
        super()._update_proxy(change)


class Abbr(Tag):
    #: Set the tag name
    tag = set_default("abbr")


class Address(Tag):
    #: Set the tag name
    tag = set_default("address")


class Cite(Tag):
    #: Set the tag name
    tag = set_default("cite")


class Bdo(Tag):
    #: Set the tag name
    tag = set_default("bdo")

    dir = d_(Str())


class Img(Tag):
    #: Set the tag name
    tag = set_default("img")

    src = d_(Str())
    width = d_(Str())
    height = d_(Str())

    @observe("src", "width", "height")
    def _update_proxy(self, change: ChangeDict):
        super()._update_proxy(change)


class Style(Tag):
    #: Set the tag name
    tag = set_default("style")


class Link(Tag):
    #: Set the tag name
    tag = set_default("link")

    type = d_(Str())
    rel = d_(Str())
    href = d_(Str())
    media = d_(Str())

    @observe("type", "rel", "href", "media")
    def _update_proxy(self, change: ChangeDict):
        super()._update_proxy(change)


class Map(Tag):
    #: Set the tag name
    tag = set_default("map")

    name = d_(Str())

    @observe("name")
    def _update_proxy(self, change: ChangeDict):
        super()._update_proxy(change)


class Area(Tag):
    #: Set the tag name
    tag = set_default("area")

    shape = d_(Str())
    coords = d_(Str())
    href = d_(Str())

    @observe("shape", "coords", "href")
    def _update_proxy(self, change: ChangeDict):
        super()._update_proxy(change)


class Table(Tag):
    #: Set the tag name
    tag = set_default("table")


class Tr(Tag):
    #: Set the tag name
    tag = set_default("tr")


class Td(Tag):
    #: Set the tag name
    tag = set_default("td")

    colspan = d_(Str())
    rowspan = d_(Str())

    @observe("colspan", "rowspan")
    def _update_proxy(self, change: ChangeDict):
        super()._update_proxy(change)


class Th(Tag):
    #: Set the tag name
    tag = set_default("th")

    colspan = d_(Str())
    rowspan = d_(Str())

    @observe("colspan", "rowspan")
    def _update_proxy(self, change: ChangeDict):
        super()._update_proxy(change)


class THead(Tag):
    #: Set the tag name
    tag = set_default("thead")


class TBody(Tag):
    #: Set the tag name
    tag = set_default("tbody")


class TFoot(Tag):
    #: Set the tag name
    tag = set_default("tfoot")


class Caption(Tag):
    #: Set the tag name
    tag = set_default("caption")


class Ul(Tag):
    #: Set the tag name
    tag = set_default("ul")


class Ol(Tag):
    #: Set the tag name
    tag = set_default("ol")

    type = d_(Enum("", "1", "A", "a", "I", "i"))

    @observe("type")
    def _update_proxy(self, change: ChangeDict):
        super()._update_proxy(change)


class Li(Tag):
    #: Set the tag name
    tag = set_default("li")


class Dl(Tag):
    #: Set the tag name
    tag = set_default("dl")


class Dt(Tag):
    #: Set the tag name
    tag = set_default("dt")


class Dd(Tag):
    #: Set the tag name
    tag = set_default("dd")


class IFrame(Tag):
    #: Set the tag name
    tag = set_default("iframe")

    src = d_(Str())
    height = d_(Str())
    width = d_(Str())
    target = d_(Str())

    @observe("src", "height", "width", "target")
    def _update_proxy(self, change: ChangeDict):
        super()._update_proxy(change)


class Script(Tag):
    #: Set the tag name
    tag = set_default("script")
    src = d_(Str())
    type = d_(Str())

    @observe("type", "src")
    def _update_proxy(self, change: ChangeDict):
        super()._update_proxy(change)


class NoScript(Tag):
    #: Set the tag name
    tag = set_default("noscript")


class Meta(Tag):
    #: Set the tag name
    tag = set_default("meta")

    name = d_(Str())
    content = d_(Str())

    @observe("name", "content")
    def _update_proxy(self, change: ChangeDict):
        super()._update_proxy(change)


class Base(Tag):
    #: Set the tag name
    tag = set_default("base")

    href = d_(Str())
    target = d_(Enum("", "_blank", "_self", "_parent", "_top", "framename"))

    @observe("href", "target")
    def _update_proxy(self, change: ChangeDict):
        super()._update_proxy(change)


class Header(Tag):
    #: Set the tag name
    tag = set_default("header")


class Nav(Tag):
    #: Set the tag name
    tag = set_default("nav")


class Section(Tag):
    #: Set the tag name
    tag = set_default("section")


class Aside(Tag):
    #: Set the tag name
    tag = set_default("aside")


class Article(Tag):
    #: Set the tag name
    tag = set_default("article")


class Footer(Tag):
    #: Set the tag name
    tag = set_default("footer")


class Summary(Tag):
    #: Set the tag name
    tag = set_default("summary")


class Details(Tag):
    #: Set the tag name
    tag = set_default("details")


class Form(Tag):
    #: Set the tag name
    tag = set_default("form")

    action = d_(Str())
    method = d_(Enum("post", "get"))

    @observe("action", "method")
    def _update_proxy(self, change: ChangeDict):
        super()._update_proxy(change)


class Fieldset(Tag):
    #: Set the tag name
    tag = set_default("fieldset")


class Legend(Tag):
    #: Set the tag name
    tag = set_default("legend")


class Label(Tag):
    #: Set the tag name
    tag = set_default("label")


class Select(Tag):
    #: Set the tag name
    tag = set_default("select")

    name = d_(Str())
    value = d_(Str())

    def _default_name(self):
        return "{}".format(self.id)

    @observe("name", "value")
    def _update_proxy(self, change: ChangeDict):
        super()._update_proxy(change)


class Option(Tag):
    #: Set the tag name
    tag = set_default("option")

    value = d_(Str())
    selected = d_(Coerced(bool))

    @observe("value", "selected")
    def _update_proxy(self, change: ChangeDict):
        super()._update_proxy(change)


class OptGroup(Tag):
    #: Set the tag name
    tag = set_default("optgroup")

    label = d_(Str())
    disabled = d_(Coerced(bool))

    @observe("label", "disabled")
    def _update_proxy(self, change: ChangeDict):
        super()._update_proxy(change)


class Input(Tag):
    #: Set the tag name
    tag = set_default("input")

    name = d_(Str())
    type = d_(Str())
    placeholder = d_(Str())
    disabled = d_(Coerced(bool))
    checked = d_(Coerced(bool))
    value = d_(Value())

    def _default_name(self):
        return f"{self.id}"

    @observe("name", "type", "disabled", "checked", "value", "placeholder")
    def _update_proxy(self, change: ChangeDict):
        super()._update_proxy(change)


class Textarea(Tag):
    #: Set the tag name
    tag = set_default("textarea")

    name = d_(Str())
    rows = d_(Str())
    cols = d_(Str())

    def _default_name(self):
        return f"{self.id}"

    @observe("name", "rows", "cols")
    def _update_proxy(self, change: ChangeDict):
        super()._update_proxy(change)


class Button(Tag):
    #: Set the tag name
    tag = set_default("button")

    name = d_(Str())
    type = d_(Str())
    value = d_(Str("1"))

    def _default_name(self):
        return f"{self.id}"

    @observe("type")
    def _update_proxy(self, change: ChangeDict):
        super()._update_proxy(change)


class Video(Tag):
    #: Set the tag name
    tag = set_default("video")

    controls = d_(Coerced(bool))

    @observe("controls")
    def _update_proxy(self, change: ChangeDict):
        super()._update_proxy(change)


class Source(Tag):
    #: Set the tag name
    tag = set_default("source")

    src = d_(Str())
    type = d_(Str())

    @observe("src", "type")
    def _update_proxy(self, change: ChangeDict):
        super()._update_proxy(change)



"""
Copyright (c) 2017, Jairus Martin.

Distributed under the terms of the MIT License.

The full license is in the file LICENSE.text, distributed with this software.

Created on Apr 15, 2017

@author: jrm
"""
from __future__ import annotations

from typing import Any
from atom.api import ForwardInstance, Enum
from enaml.core.declarative import Declarative, d_

ChangeDict = dict[str, Any]


class Block(Declarative):
    """An object which dynamically insert's its children into another block's
    parent object.

    The 'Block' object is used to cleanly and easily insert it's children
    into the children of another object. The 'Object' instance assigned to the
    'block' property of the 'Block' will be parented with the parent of
    the 'Block'. Creating a 'Block' with no parent is a programming
    error.

    """

    #: The Block to which this blocks children should be inserted into
    block = d_(ForwardInstance(lambda: Block))

    #: If replace, replace all parent's children (except the block of course)
    mode = d_(Enum("replace", "append", "prepend"))

    def initialize(self):
        """A reimplemented initializer.

        This method will add the include objects to the parent of the
        include and ensure that they are initialized.

        """
        super().initialize()
        self.refresh_items()

    def refresh_items(self):
        block = self.block
        if block:
            # This block is setting the content of another block
            before = None

            # Remove the existing blocks children
            if self.mode == "replace":
                # Clear the blocks children
                for c in block.children[:]:
                    block.children.remove(c)
                    if not c.is_destroyed:
                        c.destroy()
            # Add this blocks children to the other block
            elif self.mode == "prepend" and block.children:
                before = block.children[0]

            block.insert_children(before, self.children)

        else:
            # This block is inserting it's children into it's parent
            self.parent.insert_children(self, self.children)

    def _observe_mode(self, change: ChangeDict):
        """If the mode changes. Refresh the items."""
        block = self.block
        if block and self.is_initialized and change["type"] == "update":
            if change["oldvalue"] == "replace":
                raise NotImplementedError
            for c in self.children:
                block.children.remove(c)
                c.set_parent(None)
            self.refresh_items()

    def _observe_block(self, change: ChangeDict):
        """A change handler for the 'objects' list of the Include.

        If the object is initialized objects which are removed will be
        unparented and objects which are added will be reparented. Old
        objects will be destroyed if the 'destroy_old' flag is True.

        """
        if self.is_initialized and change["type"] == "update":
            old_block = change["oldvalue"]
            for c in self.children:
                old_block.children.remove(c)
                c.set_parent(None)
            self.refresh_items()

    def _observe__children(self, change: ChangeDict):
        """When the children of the block change. Update the referenced
        block.

        """
        if not self.is_initialized or change["type"] != "update":
            return

        block = self.block
        new_children = change["value"]
        old_children = change["oldvalue"]
        for c in old_children:
            if c not in new_children and not c.is_destroyed:
                c.destroy()
            else:
                c.set_parent(None)

        if block:
            # This block is inserting into another block
            before = None
            if self.mode == "replace":
                block.children = []
            if self.mode == "prepend" and block.children:
                before = block.children[0]
            block.insert_children(before, new_children)
        else:
            # This block is a placeholder
            self.parent.insert_children(self, new_children)



"""
Copyright (c) 2017, Jairus Martin.

Distributed under the terms of the MIT License.

The full license is in the file LICENSE.text, distributed with this software.

Created on Apr 12, 2017

@author: jrm
"""
import inspect
from web.components import html


def generic_factory():
    from .lxml_toolkit_object import WebComponent

    return WebComponent


def html_factory():
    from .lxml_toolkit_object import RootWebComponent

    return RootWebComponent


def code_factory():
    from .lxml_code import CodeComponent

    return CodeComponent


def markdown_factory():
    from .lxml_md import MarkdownComponent

    return MarkdownComponent


def notebook_factory():
    from .lxml_ipynb import NotebookComponent

    return NotebookComponent


def raw_factory():
    from .lxml_raw import RawComponent

    return RawComponent


#: Create generic html factories
FACTORIES = {
    name: generic_factory
    for name, obj in inspect.getmembers(html)
    if inspect.isclass(obj)
}

#: Create special widgets
FACTORIES.update(
    {
        "Code": code_factory,
        "Html": html_factory,
        "Markdown": markdown_factory,
        "Notebook": notebook_factory,
        "Raw": raw_factory,
    }
)



"""
Copyright (c) 2017, Jairus Martin.

Distributed under the terms of the MIT License.

The full license is in the file LICENSE.text, distributed with this software.

Created on Aug 2, 2017

@author: jrm
"""


from atom.api import Str, Typed, ForwardTyped, observe
from enaml.core.declarative import d_
from .raw import Raw, ProxyRawNode, ChangeDict


class ProxyCode(ProxyRawNode):
    #: Reference to the declaration
    declaration = ForwardTyped(lambda: Code)

    def set_language(self, language: str):
        raise NotImplementedError

    def set_highlight_style(self, style: str):
        raise NotImplementedError


class Code(Raw):
    """A block for rendering highlighted code.
    Note: You must include the proper style/stylesheets for highlighting.
          for example: https://github.com/richleland/pygments-css
    """

    #: Reference to the proxy
    proxy = Typed(ProxyCode)

    #: Language to parse if none is given it will guess which to use
    language = d_(Str())

    #: Highlighter style to use
    highlight_style = d_(Str()).tag(attr=False)

    @observe("language", "highlight_style")
    def _update_proxy(self, change: ChangeDict):
        """The superclass implementation is sufficient."""
        super()._update_proxy(change)



"""
Copyright (c) 2017, Jairus Martin.

Distributed under the terms of the MIT License.

The full license is in the file LICENSE.text, distributed with this software.

Created on Aug 2, 2017

@author: jrm
"""
from __future__ import annotations

from typing import Union, Optional
from atom.api import ForwardTyped, Instance, Typed, set_default, observe
from lxml.etree import _Element as Element
from enaml.core.declarative import d_
from .html import Tag, ProxyTag, ChangeDict

SourceType = Optional[Union[str, list[Element], Element]]


class ProxyRawNode(ProxyTag):
    #: Reference to the declaration
    declaration = ForwardTyped(lambda: Raw)

    def set_source(self, source: SourceType):
        raise NotImplementedError


class Raw(Tag):
    """A block that renders it's text as HTML.

    Note: This will ERASE any child elements the the source content!

    """

    #: Reference to the proxy
    proxy = Typed(ProxyRawNode)

    #: Default tag is a div
    tag = set_default("div")

    #: Raw source to parse and display
    source = d_(Instance((str, list, Element))).tag(attr=False)

    @observe("source")
    def _update_proxy(self, change: ChangeDict):
        """The superclass implementation is sufficient."""
        super()._update_proxy(change)



"""
Copyright (c) 2017, Jairus Martin.

Distributed under the terms of the MIT License.

The full license is in the file LICENSE.text, distributed with this software.

Created on Aug 2, 2017

@author: jrm
"""
from typing import Optional, Union
from lxml.etree import HTML
from lxml.etree import _Element as Element
from web.components.raw import ProxyRawNode, SourceType
from .lxml_toolkit_object import WebComponent


class RawComponent(WebComponent, ProxyRawNode):
    """A block for rendering raw html source."""

    def init_widget(self):
        """Initialize the widget with the source."""
        d = self.declaration
        source = d.source
        if source is not None and len(source):
            self.set_source(source)
        else:
            super().init_widget()

    def set_source(self, source: SourceType):
        """Set the source by parsing the source and inserting it into the
        component.
        """
        widget = self.widget
        assert widget is not None
        widget.clear()
        if isinstance(source, Element):
            widget.append(source)
        elif isinstance(source, list):
            widget.extend(source)
        elif isinstance(source, str):
            html = HTML(source)
            widget.extend(html[0])
        # else source is None

        # Clear removes everything so it must be reinitialized
        super().init_widget()



"""
Copyright (c) 2018, Jairus Martin.

Distributed under the terms of the MIT License.

The full license is in the file LICENSE.text, distributed with this software.

Created on Jul 8, 2018

@author: jrm
"""
from __future__ import annotations

from atom.api import Int, Typed, ForwardTyped, observe
from enaml.core.declarative import d_
from .raw import Raw, ProxyRawNode, ChangeDict


class ProxyNotebook(ProxyRawNode):
    #: Reference to the declaration
    declaration = ForwardTyped(lambda: Notebook)

    def set_version(self, version: int):
        raise NotImplementedError


class Notebook(Raw):
    """A node for rendering jupyter notebooks."""

    #: Reference to the proxy
    proxy = Typed(ProxyNotebook)

    #: Version
    version = d_(Int(4))

    @observe("version")
    def _update_proxy(self, change: ChangeDict):
        """Update the version"""
        super()._update_proxy(change)



"""
Copyright (c) 2017, Jairus Martin.

Distributed under the terms of the MIT License.

The full license is in the file LICENSE.text, distributed with this software.

Created on Aug 2, 2017

@author: jrm
"""
from __future__ import annotations

import markdown
from web.components.md import ProxyMarkdown
from .lxml_raw import RawComponent, SourceType


class MarkdownComponent(RawComponent, ProxyMarkdown):
    """A block for rendering Markdown source."""

    def _refresh_source(self):
        d = self.declaration
        assert d is not None
        source = d.source
        if isinstance(source, str):
            source = markdown.markdown(
                source,
                tab_length=d.tab_length,
                safe_mode=d.safe_mode,
                output_format=d.output_format,
                extensions=d.extensions,
                extension_configs=d.extension_configs,
            )

        #: Parse md and put in a root node
        super().set_source(source)

    def set_source(self, source: SourceType):
        self._refresh_source()

    def set_safe_mode(self, mode: bool):
        self._refresh_source()

    def set_output_format(self, format: str):
        self._refresh_source()

    def set_tab_length(self, length: int):
        self._refresh_source()

    def set_extensions(self, extensions: list[str]):
        self._refresh_source()

    def set_extension_configs(self, config: dict[str, dict]):
        self._refresh_source()



"""
Copyright (c) 2018, Jairus Martin.

Distributed under the terms of the MIT License.

The full license is in the file LICENSE.text, distributed with this software.

Created on Jul 8, 2018

@author: jrm
"""
import nbformat
from atom.api import Instance, Value
from nbconvert import HTMLExporter
from web.components.ipynb import ProxyNotebook
from .lxml_raw import RawComponent, SourceType


class NotebookComponent(RawComponent, ProxyNotebook):
    """A component for rendering Jupyter Notebooks."""

    #: Exporter
    exporter = Instance(HTMLExporter, ())

    #: Resources from the node
    resources = Value()

    def set_source(self, source: SourceType):
        # Parse md and put in a root node
        d = self.declaration
        assert d is not None
        source = d.source
        if isinstance(source, str):
            source, self.resources = self.exporter.from_notebook_node(
                nbformat.reads(source, as_version=d.version)
            )
            source = f"<div>{source}</div>"
        super().set_source(source)

    def set_version(self, version: int):
        d = self.declaration
        assert d is not None
        self.set_source(d.source)



"""
Copyright (c) 2017, Jairus Martin.

Distributed under the terms of the MIT License.

The full license is in the file LICENSE.txt, distributed with this software.

Created on Apr 12, 2017

@author: jrm
"""
from __future__ import annotations

from functools import lru_cache
from typing import Any, Type, Union, Optional, Generator
from atom.api import Atom, Bool, Member, Typed, Event, ForwardTyped, Dict, atomref
from lxml.html import tostring
from lxml.etree import _Element, Element, SubElement
from web.components.html import ProxyTag, Tag
from web.core.app import WebApplication


@lru_cache(1024)
def get_fields(cls: Type[Atom]) -> tuple[Member, ...]:
    """Determine the list of attributes to convert to html and cache them.

    Any member tagged with "attr=False" will be ignored any enaml attrs will
    also be ignored.

    """
    web_members = []
    for name, member in cls.members().items():
        meta = member.metadata
        if not meta:
            continue

        # Exclude any attr tags
        if not (meta.get("d_member") and meta.get("d_final")):
            continue

        # Skip any items tagged with attr=false
        elif not meta.get("attr", True):
            continue

        elif isinstance(member, Event):
            continue
        web_members.append(member)
    return tuple(web_members)


class WebComponent(ProxyTag):
    """An lxml implementation of an Enaml ProxyToolkitObject."""

    #: A reference to the toolkit widget created by the proxy.
    widget = Typed(_Element)

    # -------------------------------------------------------------------------
    # Initialization API
    # -------------------------------------------------------------------------
    def create_widget(self):
        """Create the toolkit widget for the proxy object.

        This method is called during the top-down pass, just before the
        'init_widget()' method is called. This method should create the
        toolkit widget and assign it to the 'widget' attribute.

        """
        d = self.declaration
        parent = d.parent.proxy
        self.root = parent.root
        self.widget = SubElement(parent.widget, d.tag)

    def init_widget(self):
        """Initialize the state of the toolkit widget.

        This method is called during the top-down pass, just after the
        'create_widget()' method is called. This method should init the
        state of the widget. The child widgets will not yet be created.

        """
        widget = self.widget
        d = self.declaration

        #: Save ref id
        self.root.cache[d.id] = atomref(self)
        attrib = widget.attrib
        attrib["id"] = d.id
        if d.text:
            widget.text = d.text
        if d.tail:
            widget.tail = d.tail
        if d.alt:
            attrib["alt"] = d.alt
        if d.style:
            self.set_style(d.style)
        if d.cls:
            self.set_cls(d.cls)
        if d.attrs:
            attrib.update(d.attrs)
        if d.clickable:
            attrib["clickable"] = "true"
        if d.draggable:
            attrib["draggable"] = "true"
        if d.onclick:
            attrib["onclick"] = d.onclick
        if d.ondragstart:
            attrib["ondragstart"] = d.ondragstart
        if d.ondragover:
            attrib["ondragover"] = d.ondragover
        if d.ondragend:
            attrib["ondragend"] = d.ondragend
        if d.ondragenter:
            attrib["ondragenter"] = d.ondragenter
        if d.ondragleave:
            attrib["ondragleave"] = d.ondragleave
        if d.ondrop:
            attrib["ondrop"] = d.ondrop

        for m in get_fields(d.__class__):
            name = m.name
            value = getattr(d, name)
            if value is True:
                attrib[name] = name
            elif value:
                attrib[name] = f"{value}"

    # -------------------------------------------------------------------------
    # ProxyToolkitObject API
    # -------------------------------------------------------------------------
    def activate_top_down(self):
        """Activate the proxy for the top-down pass."""
        self.create_widget()
        self.init_widget()

    def destroy(self):
        """A reimplemented destructor.

        This destructor will clear the reference to the toolkit widget
        and set its parent to None.

        """
        widget = self.widget
        if widget is not None:
            parent = widget.getparent()
            if parent is not None:
                parent.remove(widget)
            del self.widget

            # Remove from cache
            self.root.cache.pop(self.declaration.id, None)

        super().destroy()

    def child_added(self, child: WebComponent):
        """Handle the child added event from the declaration.

        This handler will insert the child toolkit widget in the correct.
        position. Subclasses which need more control should reimplement this
        method.

        """
        # Use insert to put in the correct spot
        d = self.declaration
        assert d is not None
        i = d.children.index(child.declaration)
        widget = self.widget
        if widget is None:
            return False
        widget.insert(i, child.widget)

    def child_moved(self, child: WebComponent):
        """Handle the child moved event from the declaration.

        This handler will pop the child and insert it in the correct position
        if it isn't already there.

        Subclasses which need more control should reimplement this method.

        Returns
        -------
        was_moved: Bool
            Whether a move was performed or not

        """
        # Determine the new index
        d = self.declaration
        assert d is not None
        i = d.children.index(child.declaration)
        w = self.widget
        if w is None:
            return False
        j = w.index(child.widget)
        if j == i:
            return False  # Already in the correct spot
        # Delete and re-insert at correct position
        del w[j]
        w.insert(i, child.widget)
        return True

    def child_removed(self, child: WebComponent):
        """Handle the child removed event from the declaration.

        This handler will unparent the child toolkit widget. Subclasses
        which need more control should reimplement this method.

        """
        w = self.widget
        if w is not None:
            w.remove(child.widget)

    # -------------------------------------------------------------------------
    # Public API
    # -------------------------------------------------------------------------
    def render(self, method: str = "html", encoding: str = "unicode", **kwargs) -> str:
        """Render the widget tree into a string"""
        return tostring(self.widget, method=method, encoding=encoding, **kwargs)

    def xpath(self, query: str, **kwargs) -> Generator[WebComponent, None, None]:
        """Get the node(s) matching the query"""
        w = self.widget
        if w is None:
            return None
        nodes = w.xpath(query, **kwargs)
        if not nodes:
            return None

        root = self.root
        assert root is not None
        lookup = root.cache.get
        for node in nodes:
            aref = lookup(node.attrib.get("id"))
            obj = aref() if aref else None
            if obj is None:
                continue
            yield obj

    def parent_widget(self) -> Optional[_Element]:
        """Get the parent toolkit widget for this object.

        Returns
        -------
        result : QObject or None
            The toolkit widget declared on the declaration parent, or
            None if there is no such parent.

        """
        parent = self.parent()
        if parent is not None:
            return parent.widget
        return None

    def child_widgets(self) -> Generator[_Element, None, None]:
        """Get the child toolkit widgets for this object.

        Returns
        -------
        result : iterable of QObject
            The child widgets defined for this object.

        """
        for child in self.children():
            w = child.widget
            if w is not None:
                yield w

    # -------------------------------------------------------------------------
    # Change handlers
    # -------------------------------------------------------------------------
    def set_text(self, text: str):
        w = self.widget
        assert w is not None
        w.text = text

    def set_tail(self, text: str):
        w = self.widget
        assert w is not None
        w.tail = text

    def set_tag(self, tag: str):
        w = self.widget
        assert w is not None
        w.tag = tag

    def set_attrs(self, attrs: dict[str, Any]):
        """Set any attributes not explicitly defined"""
        w = self.widget
        assert w is not None
        w.attrib.update(attrs)

    def set_cls(self, cls: Union[tuple[str], list[str], str]):
        if isinstance(cls, (tuple, list)):
            cls = " ".join(cls)
        w = self.widget
        assert w is not None
        w.attrib["class"] = cls

    def set_style(self, style: Union[dict, str]):
        if isinstance(style, dict):
            style = ";".join(f"{k}:{v}" for k, v in style.items())
        w = self.widget
        assert w is not None
        w.attrib["style"] = style

    def set_attribute(self, name: str, value: Any):
        """Default handler for those not explicitly defined"""
        w = self.widget
        assert w is not None
        if value is True:
            w.attrib[name] = name
        elif value is False:
            del w.attrib[name]
        else:
            w.attrib[name] = f"{value}"

    def set_clickable(self, clickable: bool):
        w = self.widget
        assert w is not None
        w.attrib["clickable"] = "true" if clickable else "false"

    def set_draggable(self, draggable: bool):
        """The draggable attr must be explicitly set to true or false"""
        w = self.widget
        assert w is not None
        w.attrib["draggable"] = "true" if draggable else "false"


class RootWebComponent(WebComponent):
    """A root component"""

    #: Components are cached for lookup by id so xpath queries from lxml
    #: can retrieve their declaration component
    cache = Dict()

    #: Flag to indicate whether this node was rendered. This is used by the
    #: declaration to avoid creating unnecessary modified events.
    rendered = Bool()

    def create_widget(self):
        self.root = self  # This is the root
        self.widget = Element(self.declaration.tag)

    def render(self, *args, **kwargs):
        self.rendered = True
        return super().render(*args, **kwargs)



"""
Copyright (c) 2017, Jairus Martin.

Distributed under the terms of the MIT License.

The full license is in the file LICENSE.text, distributed with this software.

Created on Aug 2, 2017

@author: jrm
"""

from typing import Optional
from atom.api import Typed, ForwardTyped, Enum, Int, Bool, List, Dict, observe
from enaml.core.declarative import d_
from .html import Tag, ChangeDict
from .raw import Raw, ProxyRawNode


class ProxyMarkdown(ProxyRawNode):
    #: Reference to the declaration
    declaration = ForwardTyped(lambda: Markdown)

    def set_safe_mode(self, mode: bool):
        raise NotImplementedError

    def set_output_format(self, output_format: str):
        raise NotImplementedError

    def set_tab_length(self, length: int):
        raise NotImplementedError

    def set_extensions(self, extensions: list[str]):
        raise NotImplementedError

    def set_extensions_config(self, config: dict[str, dict]):
        raise NotImplementedError


class Markdown(Raw):
    """A block for rendering Markdown source."""

    #: Extensions to use when rendering
    extensions = d_(
        List(
            default=[
                "markdown.extensions.codehilite",
                "markdown.extensions.fenced_code",
                "markdown.extensions.tables",
            ]
        )
    ).tag(attr=False)

    #: Configuration for them
    extension_configs = d_(
        Dict(
            default={
                "markdown.extensions.codehilite": {"css_class": "highlight"},
            }
        )
    ).tag(attr=False)

    #: Disallow raw HTMl
    safe_mode = d_(Bool()).tag(attr=False)

    #: Output format
    output_format = d_(Enum("xhtml", "html5")).tag(attr=False)

    #: Tab size
    tab_length = d_(Int(4)).tag(attr=False)

    #: Reference to the proxy
    proxy = Typed(ProxyMarkdown)

    @observe(
        "extensions", "extension_configs", "safe_mode", "output_format", "tab_length"
    )
    def _update_proxy(self, change: ChangeDict):
        """The superclass implementation is sufficient."""
        super()._update_proxy(change)

    def _notify_modified(self, root: Optional[Tag], change: ChangeDict):
        """Update the notification"""
        if change["type"] == "update" and change["name"] == "source":
            change["value"] = self.render()
        super()._notify_modified(root, change)



"""
Created on Apr 17, 2017

@author: jrm
"""
from setuptools import Extension, setup, find_packages

speedups_module = Extension(
    "web.core.speedups",
    ["src/speedups.c"],
    include_dirs=["src"],
    language="c",
)

setup(
    name='enaml-web',
    version='0.11.1',
    author='CodeLV',
    author_email='frmdstryr@gmail.com',
    url='https://github.com/codelv/enaml-web',
    description='Web component toolkit for Enaml',
    license="MIT",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    requires=['enaml'],
    python_requires='>=3.9',
    install_requires=['enaml >= 0.9.8', 'lxml>=3.4.0'],
    optional_requires=[
        'Pygments', 'Markdown', 'nbconvert',  # extra components
    ],
    ext_modules=[speedups_module],
    packages=find_packages(),
)
